"""Constants and built-in language maps for sluggi."""

import re

# Regex pattern strings for emoji and word extraction
_RE_EMOJI_ADJACENT = re.compile(r"(:[a-zA-Z0-9_\-&]+:)(?=:[a-zA-Z0-9_\-&]+:)")
_RE_EMOJI_NAME = re.compile(r":([a-zA-Z0-9_\-&]+):")
_RE_NON_WORD = re.compile(r"[^\w\-]+", re.UNICODE)
_RE_SEPARATOR_MULTI = re.compile(r"[-_\.]{2,}")
_RE_WORD = re.compile(r"\w+")

# Regex for HTML entity and numeric reference decoding
_RE_HTML_ENTITY_DECIMAL = re.compile(r"&#(\d+);")
_RE_HTML_ENTITY_HEXADECIMAL = re.compile(r"&#x([0-9A-Fa-f]+);")

# Built-in transliteration map: covers Cyrillic, Greek, and basic Latinization
_BUILTIN_TRANSLIT_MAP = {
    # IMPORTANT: For correct replacement order, keys with longer length should appear
    # earlier in this mapping.
    # Special-case Greek greeting
    "γεια": "geia",
    "ΓΕΙΑ": "GEIA",
    # Multi-letter Greek digraphs
    "ια": "ia",
    "ΙΑ": "IA",
    # Greek (partial, unambiguous Latin equivalents)
    "Α": "A",
    "Β": "B",
    "Γ": "G",
    "Δ": "D",
    "Ε": "E",
    "Ζ": "Z",
    "Η": "I",
    "Θ": "Th",
    "Ι": "I",
    "Κ": "K",
    "Λ": "L",
    "Μ": "M",
    "Ν": "N",
    "Ξ": "X",
    "Ο": "O",
    "Π": "P",
    "Ρ": "R",
    "Σ": "S",
    "Τ": "T",
    "Υ": "Y",
    "Φ": "F",
    "Χ": "X",
    "α": "a",
    "β": "b",
    "γ": "g",
    "δ": "d",
    "ε": "e",
    "ζ": "z",
    "η": "i",
    "θ": "th",
    "ι": "i",
    "κ": "k",
    "λ": "l",
    "μ": "m",
    "ν": "n",
    "ξ": "x",
    "ο": "o",
    "π": "p",
    "ρ": "r",
    "σ": "s",
    "ς": "s",
    "τ": "t",
    "υ": "u",
    "φ": "f",
    "χ": "x",
    "ψ": "ps",
    "ω": "o",
    # Greek diacritics (map directly to Latin equivalents)
    "ά": "a",
    "έ": "e",
    "ή": "i",
    "ί": "i",
    "ό": "o",
    "ύ": "u",
    "ώ": "o",
    "Ά": "A",
    "Έ": "E",
    "Ή": "I",
    "Ί": "I",
    "Ό": "O",
    "Ύ": "U",
    "Ώ": "O",
    # Greek digraphs/phonemes (pre-processing for better transliteration)
    "Ψ": "Ps",
    "Ω": "O",  # Cyrillic (partial, unambiguous Latin equivalents)
    "А": "A",
    "Б": "B",
    "В": "V",
    "Г": "G",
    "Д": "D",
    "Е": "E",
    "Ё": "E",
    "Ж": "Zh",
    "З": "Z",
    "И": "I",
    "Й": "I",
    "К": "K",
    "Л": "L",
    "М": "M",
    "Н": "N",
    "О": "O",
    "П": "P",
    "Р": "R",
    "С": "S",
    "Т": "T",
    "У": "U",
    "Ф": "F",
    "Х": "Kh",
    "Ц": "Ts",
    "Ч": "Ch",
    "Ш": "Sh",
    "Щ": "Shch",
    "Ы": "Y",
    "Э": "E",
    "Ю": "Yu",
    "Я": "Ya",
    "а": "a",
    "б": "b",
    "в": "v",
    "г": "g",
    "д": "d",
    "е": "e",
    "ё": "e",
    "ж": "zh",
    "з": "z",
    "и": "i",
    "й": "i",
    "к": "k",
    "л": "l",
    "м": "m",
    "н": "n",
    "о": "o",
    "п": "p",
    "р": "r",
    "с": "s",
    "т": "t",
    "у": "u",
    "ф": "f",
    "х": "kh",
    "ц": "ts",
    "ч": "ch",
    "ш": "sh",
    "щ": "shch",
    "ы": "y",
    "э": "e",
    "ю": "yu",
    "я": "ya",
}

_TRANSLIT_REGEX = re.compile("|".join(map(re.escape, _BUILTIN_TRANSLIT_MAP.keys())))
